from __future__ import annotations

import io
import logging
import os
import sys
import time
import weakref
from collections.abc import Iterable
from typing import TYPE_CHECKING

import gams.transfer as gt
import pandas as pd
from gams.core.cfg import GMS_SSSIZE
from gams.core.gev import (
    gevCreateD,
    gevFree,
    gevGetLShandle,
    gevGetStrOpt,
    gevHandleToPtr,
    gevInitEnvironmentLegacy,
    gevNameLogFile,
    gevNameStaFile,
    gevRestoreLogStat,
    gevRestoreLogStatRewrite,
    gevSwitchLogStat,
    new_gevHandle_tp,
)
from gams.core.gmd import (
    GMD_NRUELS,
    gmdCallSolver,
    gmdCloseLicenseSession,
    gmdInfo,
    gmdInitFromDict,
    gmdInitUpdate,
    gmdUpdateModelSymbol,
)
from gams.core.gmo import (
    gmoCreateD,
    gmoFree,
    gmoGetHeadnTail,
    gmoHandleToPtr,
    gmoHdomused,
    gmoHetalg,
    gmoHiterused,
    gmoHmarginals,
    gmoHobjval,
    gmoHresused,
    gmoLoadDataLegacy,
    gmoModelStat,
    gmoNameOptFileSet,
    gmoOptFileSet,
    gmoRegisterEnvironment,
    gmoSolveStat,
    gmoTmipbest,
    gmoTmipnod,
    new_gmoHandle_tp,
)

import gamspy as gp
import gamspy._symbols.implicits as implicits
import gamspy.utils as utils
from gamspy._database import (
    Database,
    GamsEquation,
    GamsParameter,
    GamsVariable,
)
from gamspy._options import FreezeOptions, Options, write_solver_options
from gamspy.exceptions import (
    GamspyException,
    ValidationError,
    _customize_exception,
)

if TYPE_CHECKING:
    from gamspy import Container, Model, Parameter
    from gamspy._symbols.implicits import ImplicitParameter

logger = logging.getLogger("FROZEN MODEL")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


VARIABLE_MAP = {
    "binary": 1,
    "integer": 2,
    "positive": 3,
    "negative": 4,
    "free": 5,
    "sos1": 6,
    "sos2": 7,
    "semicont": 8,
    "semiint": 9,
}


EQUATION_MAP = {
    "eq": 0,
    "geq": 1,
    "leq": 2,
    "nonbinding": 3,
    "external": 4,
}

UPDATE_ACTION_MAP = {
    "up": 1,
    "lo": 2,
    "fx": 3,
    "l": 4,
    "m": 5,
}

UPDATE_TYPE_MAP = {
    "0": 0,
    "base_case": 1,
    "accumulate": 2,
    "inherit": 3,
}


class GamsModifier:
    def __init__(
        self,
        gams_symbol: GamsParameter | GamsVariable | GamsEquation,
        update_action: int | None = None,
        data_symbol: GamsParameter | None = None,
        update_type: int = 3,
    ):
        self.update_action = None
        self.gams_symbol = gams_symbol
        self.update_type = update_type
        self.data_symbol = None

        # update_action and data_symbol specified
        if update_action is not None and data_symbol is not None:
            self._validate_update_action(update_action)

            self.update_action = update_action
            self.data_symbol = data_symbol
        # only the gams_symbol is specified
        elif update_action is None and data_symbol is None:
            ...
        else:
            raise GamspyException(
                "Wrong combination of parameters. Specifying only update_action or data_symbol is not allowed."
            )

    def _validate_update_action(self, update_action: int):
        if update_action in (1, 2, 3):
            if not isinstance(self.gams_symbol, GamsVariable):
                raise GamspyException(
                    f"GAMS Symbol must be GAMSVariable for {update_action}"
                )
        elif update_action in (4, 5):
            if not (
                isinstance(self.gams_symbol, (GamsVariable, GamsEquation))
            ):
                raise GamspyException(
                    f"GAMS Symbol must be GAMSVariable or GAMSEquation for {update_action}"
                )
        else:
            raise GamspyException(f"Unknown update action {update_action}")


class ModelInstance:
    """
    ModelInstance class provides a controlled way of modifying a model instance
    and solving the resulting problem in the most efficient way, by communicating
    only the changes of the model to the solver and doing a hot start (in case of
    a continuous model like LP) without the use of disk IO.

    Parameters
    ----------
    container : Container
    model : Model
    modifiables : list[Parameter  |  ImplicitParameter]
    freeze_options : Options | dict | None, optional
    """

    def __init__(
        self,
        container: Container,
        model: Model,
        modifiables: list[Parameter | ImplicitParameter],
        freeze_options: Options,
        output: io.TextIOWrapper | None,
    ):
        self.container = container
        self.job_name = container._job
        self.gms_file = self.job_name + ".gms"
        self.lst_file = self.job_name + ".lst"
        self.pf_file = self.job_name + ".pf"
        self.trace_file = self.job_name + ".txt"
        self.solver_control_file = os.path.join(
            self.container._process_directory, "gamscntr.dat"
        )

        self.model = model
        self.output = output
        assert self.model._is_frozen

        self.modifiables = self._init_modifiables(modifiables)
        self.instance_container = gt.Container(
            system_directory=container.system_directory,
        )

        self.workspace = container._workspace
        self.sync_db = Database(self.workspace)

        self._gev = new_gevHandle_tp()
        ret = gevCreateD(self._gev, container.system_directory, GMS_SSSIZE)
        if not ret[0]:
            raise GamspyException(ret[1])

        self._gmo = new_gmoHandle_tp()
        ret = gmoCreateD(self._gmo, container.system_directory, GMS_SSSIZE)
        if not ret[0]:
            raise GamspyException(ret[1])

        self.modifiers = self._create_modifiers()
        self.instantiate(model, freeze_options)

        # preallocate summary frame for performance reasons
        HEADER = [
            "Solver Status",
            "Model Status",
            "Objective",
            "Solver",
            "Solver Time",
        ]
        self.summary = pd.DataFrame(index=range(1), columns=HEADER)

        weakref.finalize(self, self.cleanup, self._gmo, self._gev)

    @staticmethod
    def cleanup(gmo, gev) -> None:
        gmoFree(gmo)
        gevFree(gev)

    def close_license_session(self) -> None:
        gmdCloseLicenseSession(self.sync_db.gmd)

    def _create_modifiers(self) -> list[GamsModifier]:
        modifiers = []

        for symbol in self.modifiables:
            if isinstance(symbol, gp.Parameter):
                domain = ["*"] * symbol.dimension
                _ = gt.Parameter(self.instance_container, symbol.name, domain)
                modifiers.append(
                    GamsModifier(
                        self.sync_db.add_parameter(
                            symbol.name,
                            symbol.dimension,
                            symbol.description,
                        )
                    )
                )

            elif isinstance(symbol, implicits.ImplicitParameter):
                attribute = symbol.name.split(".")[-1]
                update_action = UPDATE_ACTION_MAP[attribute]

                try:
                    sync_db_symbol = self.sync_db[symbol.parent.name]
                except KeyError:
                    if isinstance(symbol.parent, gp.Variable):
                        sync_db_symbol = self.sync_db.add_variable(
                            symbol.parent.name,
                            symbol.parent.dimension,
                            VARIABLE_MAP[symbol.parent.type],
                        )

                    elif isinstance(symbol.parent, gp.Equation):
                        sync_db_symbol = self.sync_db.add_equation(
                            symbol.parent.name,
                            symbol.parent.dimension,
                            EQUATION_MAP[symbol.parent.type],
                        )

                attr_name = "_".join(symbol.name.split("."))

                domain = ["*"] * symbol.parent.dimension
                _ = gt.Parameter(
                    self.instance_container,
                    attr_name,
                    domain=domain,
                )

                data_symbol = self.sync_db.add_parameter(
                    attr_name,
                    symbol.parent.dimension,
                )
                modifiers.append(
                    GamsModifier(sync_db_symbol, update_action, data_symbol)
                )
            else:
                raise ValidationError(
                    f"Symbol type {type(symbol)} cannot be modified in a"
                    " frozen solve"
                )

        return modifiers

    def instantiate(self, model: Model, options: Options) -> None:
        # Check the gmd state.
        rc, _, _, _ = gmdInfo(self.sync_db.gmd, GMD_NRUELS)
        self.sync_db._check_for_gmd_error(rc, self.workspace)

        # Prepare the required lines to solve with model instance
        scenario_str = self._get_scenario(model)
        with open(self.gms_file, "w", encoding="utf-8") as gams_file:
            gams_file.write(scenario_str)

        # Write pf file
        extra_options = self._prepare_gams_options()
        options._set_extra_options(extra_options)
        options.log_file = os.path.join(
            self.container.working_directory, "gamslog.dat"
        )
        options._export(self.pf_file, self.output)

        # Run
        try:
            self.container._job = self.job_name
            self.container._send_job(self.job_name, self.pf_file, self.output)
        except GamspyException as exception:
            self.container._workspace._errors.append(str(exception))
            message = _customize_exception(
                options,
                self.job_name,
                exception.return_code,
            )

            exception.args = (exception.message + message,)
            raise exception
        finally:
            self.container._unsaved_statements = []

        # Init environments
        if gevInitEnvironmentLegacy(self._gev, self.solver_control_file) != 0:
            raise GamspyException("Could not initialize model instance")

        gmoRegisterEnvironment(self._gmo, gevHandleToPtr(self._gev))
        ret = gmoLoadDataLegacy(self._gmo)
        if ret[0] != 0:
            raise GamspyException(f"Could not load model instance: {ret[1]}")

        rc = gmdInitFromDict(self.sync_db.gmd, gmoHandleToPtr(self._gmo))
        self.sync_db._check_for_gmd_error(rc, self.workspace)

    def solve(
        self,
        solver: str,
        instance_options: FreezeOptions,
        solver_options: dict | None,
        output: io.TextIOWrapper | None,
    ) -> pd.DataFrame:
        # write solver options file
        option_file = 0
        if solver_options:
            write_solver_options(
                self.container.system_directory,
                self.container.working_directory,
                solver,
                solver_options,
            )
            option_file = 1

        names_to_write = []
        for symbol in self.modifiables:
            if isinstance(symbol, gp.Parameter):
                self.instance_container[symbol.name].records = self.container[
                    symbol.name
                ].records
                names_to_write.append(symbol.name)

            if (
                isinstance(symbol, implicits.ImplicitParameter)
                and symbol.parent.records is not None
            ):
                parent_name, attr = symbol.name.split(".")
                attr_name = "_".join([parent_name, attr])

                columns = self._get_columns_to_drop(attr)

                self.instance_container[attr_name].setRecords(
                    self.container[parent_name].records.drop(columns, axis=1)
                )
                names_to_write.append(attr_name)

            self.instance_container.write(
                self.sync_db.gmd, symbols=names_to_write, eps_to_zero=False
            )

        ### Legacy code from GAMS Control. TODO: Pay the technical debt of the following legacy code. ###
        rc = gmdInitUpdate(self.sync_db.gmd, gmoHandleToPtr(self._gmo))
        self.sync_db._check_for_gmd_error(rc, self.workspace)

        # Update gmd
        start = time.perf_counter()
        accumulate_no_match_cnt = 0
        no_match_cnt = 0

        for modifier in self.modifiers:
            update_type = UPDATE_TYPE_MAP[instance_options.update_type]
            if modifier.update_type != 3:
                update_type = modifier.update_type

            if isinstance(modifier.gams_symbol, GamsParameter):
                rc, no_match_cnt = gmdUpdateModelSymbol(
                    self.sync_db.gmd,
                    modifier.gams_symbol.sym_ptr,
                    0,
                    modifier.gams_symbol.sym_ptr,
                    update_type,
                    no_match_cnt,
                )
                self.sync_db._check_for_gmd_error(rc, self.workspace)
            else:
                rc, no_match_cnt = gmdUpdateModelSymbol(
                    self.sync_db.gmd,
                    modifier.gams_symbol.sym_ptr,
                    modifier.update_action,
                    modifier.data_symbol.sym_ptr,  # type: ignore
                    update_type,
                    no_match_cnt,
                )
                self.sync_db._check_for_gmd_error(rc, self.workspace)

            accumulate_no_match_cnt += no_match_cnt
            if accumulate_no_match_cnt > instance_options.no_match_limit:
                raise GamspyException(
                    f"Unmatched record limit exceeded while processing modifier {modifier.gams_symbol.name}, for more info check no_match_limit option."
                )

        model_generation_time = time.perf_counter() - start

        # Close Log and status file and remove
        if output:
            gevSwitchLogStat(
                self._gev, 0, "", False, "", False, None, None, None
            )
            ls_handle = gevGetLShandle(self._gev)
            gevRestoreLogStatRewrite(self._gev, ls_handle)

        if output == sys.stdout:
            gevSwitchLogStat(
                self._gev,
                3,
                gevGetStrOpt(self._gev, gevNameLogFile),
                False,
                gevGetStrOpt(self._gev, gevNameStaFile),
                False,
                None,
                None,
                ls_handle,
            )
            ls_handle = gevGetLShandle(self._gev)

        if instance_options is not None and instance_options.debug:
            with open(
                os.path.join(self.workspace.working_directory, "convert.opt"),
                "w",
            ) as opt_file:
                opt_file.writelines(
                    [
                        "gams "
                        + os.path.join(
                            self.workspace.working_directory, "gams.gms"
                        ),
                        "dumpgdx "
                        + os.path.join(
                            self.workspace.working_directory, "dump.gdx\n"
                        ),
                        "dictmap "
                        + os.path.join(
                            self.workspace.working_directory, "dictmap.gdx"
                        ),
                    ]
                )

                gmoOptFileSet(self._gmo, 1)
                gmoNameOptFileSet(
                    self._gmo,
                    os.path.join(
                        self.workspace.working_directory, "convert.opt"
                    ),
                )
                rc = gmdCallSolver(self.sync_db.gmd, "convert")
                self.sync_db._check_for_gmd_error(rc, self.workspace)

        gmoOptFileSet(self._gmo, option_file)
        gmoNameOptFileSet(
            self._gmo,
            os.path.join(
                self.workspace.working_directory, solver.lower() + ".opt"
            ),
        )

        rc = gmdCallSolver(self.sync_db.gmd, solver)
        self.sync_db._check_for_gmd_error(rc, self.workspace)

        if output == sys.stdout:
            gevRestoreLogStat(self._gev, ls_handle)

        if output is not None and output != sys.stdout:
            gevSwitchLogStat(
                self._gev, 0, "", False, "", False, None, None, ls_handle
            )
            ls_handle = gevGetLShandle(self._gev)
            with open(gevGetStrOpt(self._gev, gevNameLogFile)) as file:
                for line in file.readlines():
                    output.write(line)
                gevRestoreLogStat(self._gev, ls_handle)
        ### end of the legacy code ###

        self._update_main_container()

        # update model attributes
        from gamspy._model import INTERRUPT_STATUS

        self.model._status = gp.ModelStatus(gmoModelStat(self._gmo))
        self.model._solve_status = gp.SolveStatus(gmoSolveStat(self._gmo))
        if self.model._solve_status in INTERRUPT_STATUS:
            logger.warning(
                f"The solve was interrupted! Solve status: {self.model._solve_status.name}. "
                "For further information, see https://gamspy.readthedocs.io/en/latest/reference/gamspy._model.html#gamspy.SolveStatus."
            )
        self.model._model_generation_time = model_generation_time
        self.model._solve_model_time = gmoGetHeadnTail(self._gmo, gmoHresused)
        self.model._num_iterations = gmoGetHeadnTail(self._gmo, gmoHiterused)
        self.model._marginals = gmoGetHeadnTail(self._gmo, gmoHmarginals)
        self.model._algorithm_time = gmoGetHeadnTail(self._gmo, gmoHetalg)
        self.model._objective_estimation = gmoGetHeadnTail(
            self._gmo, gmoTmipbest
        )
        self.model._num_nodes_used = gmoGetHeadnTail(self._gmo, gmoTmipnod)
        self.model._num_domain_violations = gmoGetHeadnTail(
            self._gmo, gmoHdomused
        )
        self.model._objective_value = gmoGetHeadnTail(self._gmo, gmoHobjval)
        self.summary.loc[0] = [
            str(self.model._solve_status),
            str(self.model._status),
            self.model._objective_value,
            solver,
            self.model._solve_model_time,
        ]

        return self.summary

    def _get_scenario(self, model: Model) -> str:
        auto_id = "s" + utils._get_unique_name()[:5]
        params = [
            modifier.gams_symbol
            for modifier in self.modifiers
            if isinstance(modifier.gams_symbol, GamsParameter)
        ]
        lines = []
        if params:
            lines.append(f"Set {auto_id}__(*) /'s0'/;")
            for symbol in params:
                declaration = f"Parameter {auto_id}__{symbol.name}({auto_id}__"
                domain = ""
                if symbol.dimension:
                    domain = "," + ",".join("*" * symbol.dimension)
                domain += ")"

                declaration = f"{declaration}{domain};"
                lines.append(declaration)

                domain = f"({auto_id}__"

                if symbol.dimension:
                    domain += ","

                assign_str = f"{auto_id}__{symbol.name}({auto_id}__"
                if symbol.dimension:
                    assign_str += "," + ",".join(
                        [f"{auto_id}__"] * symbol.dimension
                    )

                assign_str += ") = Eps;"
                lines.append(assign_str)

            scenario = (
                f"Set {auto_id}_dict(*,*,*) / '{auto_id}__'.'scenario'.''"
            )
            for symbol in params:
                scenario += (
                    f",\n'{symbol.name}'.'param'.'{auto_id}__{symbol.name}'"
                )
            scenario += "/;"
            lines.append(scenario)

        lines.append(f"{model.name}.justScrDir = 1;")
        solve_string = model._generate_solve_string()

        if params:
            solve_string += f" scenario {auto_id}_dict"

        solve_string += ";"
        lines.append(solve_string)

        return "\n".join(lines)

    def _init_modifiables(
        self, modifiables: list[Parameter | ImplicitParameter]
    ) -> list[Parameter | ImplicitParameter]:
        if not isinstance(modifiables, Iterable):
            raise ValidationError(
                "Modifiables must be iterable (i.e. list, tuple etc.)."
            )

        if any(
            not isinstance(symbol, (gp.Parameter, implicits.ImplicitParameter))
            for symbol in modifiables
        ):
            raise ValidationError(
                "Type of a modifiable must be either Parameter or a Variable attribute (e.g. variable.up)"
            )

        symbols_in_conditions: list[str] = []
        for equation in self.model.equations:
            assert equation._definition is not None
            symbols_in_conditions += (
                equation._definition._find_symbols_in_conditions()
            )

        will_be_modified: list[Parameter | ImplicitParameter] = []
        for symbol in modifiables:
            if isinstance(symbol, implicits.ImplicitParameter):
                if symbol.parent.name in symbols_in_conditions:
                    raise ValidationError(
                        f"Modifiable symbol `{symbol.parent.name}` cannot be in a condition."
                    )

                attr_name = symbol.name.split(".")[-1]

                # If the symbol attr is fx, then modify level, lower and upper.
                if attr_name == "fx":
                    if not utils.isin(symbol.parent.l, will_be_modified):
                        will_be_modified.append(symbol.parent.l)

                    if not utils.isin(symbol.parent.lo, will_be_modified):
                        will_be_modified.append(symbol.parent.lo)

                    if not utils.isin(symbol.parent.up, will_be_modified):
                        will_be_modified.append(symbol.parent.up)
                else:
                    # if fx already added level, lower or upper, do not add again.
                    if not utils.isin(symbol, will_be_modified):
                        will_be_modified.append(symbol)
            else:
                if symbol.name in symbols_in_conditions:
                    raise ValidationError(
                        f"Modifiable symbol `{symbol.name}` cannot be in a condition."
                    )
                will_be_modified.append(symbol)

        return will_be_modified

    def _prepare_gams_options(self) -> dict:
        scrdir = self.container._process_directory
        extra_options = {
            "trace": self.trace_file,
            "input": self.gms_file,
            "output": self.lst_file,
            "optdir": self.container.working_directory,
            "sysdir": self.container.system_directory,
            "scrdir": scrdir,
            "scriptnext": os.path.join(scrdir, "gamsnext.sh"),
            "license": utils._get_license_path(
                self.container.system_directory
            ),
            "solvercntr": self.solver_control_file,
        }
        if self.container._network_license:
            extra_options["netlicense"] = os.path.join(
                self.container._process_directory, "gamslice.dat"
            )

        return extra_options

    def _get_columns_to_drop(self, attr: str) -> list[str]:
        attr_map = {
            "l": "level",
            "m": "marginal",
            "lo": "lower",
            "up": "upper",
            "scale": "scale",
        }

        columns = []
        for key, value in attr_map.items():
            if key != attr:
                columns.append(value)

        return columns

    def _update_main_container(self) -> None:
        temp = gt.Container(
            system_directory=self.container.system_directory,
        )
        temp.read(self.sync_db.gmd)

        prev_state = self.container._options.miro_protect
        for name in temp.data:
            if name in self.container.data:
                self.container._options.miro_protect = False
                self.container[name].records = temp[name].records
                self.container[name].domain_labels = self.container[
                    name
                ].domain_names

            if name in (symbol.name for symbol in self.modifiables):
                generated_var = name + "_var"
                if generated_var not in self.container.data:
                    _ = gp.Variable(
                        self.container,
                        generated_var,
                        domain=self.container[name].domain,
                        records=temp[generated_var].records,
                    )
                else:
                    self.container[generated_var]._records = temp[
                        generated_var
                    ].records

        self.container._options.miro_protect = prev_state

        if self.model._objective_variable is not None:
            self.model._objective_value = temp[
                self.model._objective_variable.name
            ].toValue()



from __future__ import annotations
import os
import platform
import subprocess
import sys
from enum import Enum

import typer

from gamspy.exceptions import GamspyException, ValidationError

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To run your model with GAMS MIRO.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy run miro [--path <path_to_miro>] [--model <path_to_model>]",
    context_settings={"help_option_names": ["-h", "--help"]},
)

class ModeEnum(Enum):
    config = "config"
    base = "base"
    deploy = "deploy"

@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy run miro [--path <path_to_miro>] [--model <path_to_model>]", 
    short_help="Runs a GAMSPY model with GAMS MIRO app."
)
def miro(
    model: str = typer.Option(
        None,
        "--model",
        "-g",
        help="Path to the GAMSPy model."
    ),
    mode: ModeEnum = typer.Option(
        "base",
        "--mode",
        "-m",
        help="Execution mode of MIRO"
    ),
    path: str = typer.Option(
        None,
        "--path",
        "-p",
        help="Path to the MIRO executable (.exe on Windows, .app on macOS or .AppImage on Linux"
    ),
    skip_execution: bool = typer.Option(
        False,
        "--skip-execution",
        help="Whether to skip model execution."
    )
) -> None:
    if model is None:
        raise ValidationError("--model must be provided to run MIRO")
    
    model = os.path.abspath(model)
    execution_mode = mode.value
    path = os.getenv("MIRO_PATH", None)

    if path is None:
        path = path if path is not None else discover_miro()

    if path is None:
        raise GamspyException(
            "--path must be provided to run MIRO"
        )

    if (
        platform.system() == "Darwin"
        and os.path.splitext(path)[1] == ".app"
    ):
        path = os.path.join(path, "Contents", "MacOS", "GAMS MIRO")

    # Initialize MIRO
    if not skip_execution:
        subprocess_env = os.environ.copy()
        subprocess_env["MIRO"] = "1"

        try:
            subprocess.run(
                [sys.executable, model], env=subprocess_env, check=True
            )
        except subprocess.CalledProcessError:
            return

    # Run MIRO
    subprocess_env = os.environ.copy()
    if execution_mode == "deploy":
        subprocess_env["MIRO_BUILD"] = "true"
        execution_mode = "base"

    subprocess_env["MIRO_MODEL_PATH"] = model
    subprocess_env["MIRO_MODE"] = execution_mode
    subprocess_env["MIRO_DEV_MODE"] = "true"
    subprocess_env["MIRO_USE_TMP"] = "false"
    subprocess_env["PYTHON_EXEC_PATH"] = sys.executable

    subprocess.run([path], env=subprocess_env, check=True)

def discover_miro():
    system = platform.system()
    if system == "Linux":
        return None

    home = os.path.expanduser("~")
    standard_locations = {
        "Darwin": [
            os.path.join(
                "/",
                "Applications",
                "GAMS MIRO.app",
                "Contents",
                "MacOS",
                "GAMS MIRO",
            ),
            os.path.join(
                home,
                "Applications",
                "GAMS MIRO.app",
                "Contents",
                "MacOS",
                "GAMS MIRO",
            ),
        ],
        "Windows": [
            os.path.join(
                "C:\\", "Program Files", "GAMS MIRO", "GAMS MIRO.exe"
            ),
            os.path.join(
                home,
                "AppData",
                "Local",
                "Programs",
                "GAMS MIRO",
                "GAMS MIRO.exe",
            ),
        ],
    }

    if system in ("Darwin", "Windows"):
        for location in standard_locations[system]:
            if os.path.isfile(location):
                return location

    return None

if __name__ == "__main__":
    app()



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._symbols.implicits.implicit_symbol import ImplicitSymbol
from gamspy._types import EllipsisType
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    import pandas as pd

    from gamspy import Set, Variable
    from gamspy._algebra.expression import Expression


class ImplicitVariable(ImplicitSymbol, operable.Operable):
    def __init__(
        self,
        parent: Variable,
        name: str,
        domain: list[Set | str],
        permutation: list[int] | None = None,
        scalar_domains: list[tuple[int | Set]] | None = None,
    ):
        """
        Implicit Variable

        Parameters
        ----------
        parent : Variable
        name : str
        domain : list[Set | str]
        """
        super().__init__(parent, name, domain, permutation, scalar_domains)

    def __repr__(self) -> str:
        return f"ImplicitVariable(parent={self.parent}, name='{self.name}', domain={self.domain}, permutation={self.permutation}, parent_scalar_domains={self.parent_scalar_domains})"

    def __getitem__(
        self, indices: Sequence | str | EllipsisType | slice
    ) -> ImplicitVariable:
        domain = validation.validate_domain(self, indices)
        return ImplicitVariable(
            parent=self.parent,
            name=self.name,
            domain=domain,
            permutation=self.permutation,
            scalar_domains=self._scalar_domains,
        )

    @property
    def l(self):  # noqa: E741,E743
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.l.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @l.setter
    def l(self, value: int | float | Expression):
        # b[t].l = 30 -> b.l[t] = 30
        self.parent.l[self.domain] = value

    @property
    def m(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.m.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @m.setter
    def m(self, value: int | float | Expression):
        # b[t].m = 30 -> b.m[t] = 30
        self.parent.m[self.domain] = value

    @property
    def lo(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.lo.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @lo.setter
    def lo(self, value: int | float | Expression):
        # b[t].lo = 30 -> b.lo[t] = 30
        self.parent.lo[self.domain] = value

    @property
    def up(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.up.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @up.setter
    def up(self, value: int | float | Expression):
        # b[t].up = 30 -> b.up[t] = 30
        self.parent.up[self.domain] = value

    @property
    def scale(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.scale.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @scale.setter
    def scale(self, value: int | float | Expression):
        if self.parent.type in ("integer", "binary"):
            raise ValidationError(
                "Scales cannot be applied to discrete variables."
            )

        # b[t].scale = 30 -> b.scale[t] = 30
        self.parent.scale[self.domain] = value

    @property
    def fx(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.fx.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @fx.setter
    def fx(self, value: int | float | Expression):
        # b[t].fx = 30 -> b.fx[t] = 30
        self.parent.fx[self.domain] = value

    @property
    def prior(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.prior.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @prior.setter
    def prior(self, value: int | float | Expression):
        if self.parent.type not in ("integer", "binary"):
            raise ValidationError(
                "Priorities can only be used on discrete variables."
            )

        # b[t].prior = 30 -> b.prior[t] = 30
        self.parent.prior[self.domain] = value

    @property
    def stage(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.stage.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @stage.setter
    def stage(self, value: int | float | Expression):
        # b[t].stage = 30 -> b.stage[t] = 30
        self.parent.stage[self.domain] = value

    @property
    def T(self) -> implicits.ImplicitVariable:
        """See gamspy.ImplicitVariable.t"""
        return self.t()

    def t(self) -> implicits.ImplicitVariable:
        """Returns an ImplicitVariable derived from this
        implicit variable by swapping its last two indices.
        This operation does not generate a new variable in GAMS.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> j = gp.Set(m, "j", records=['j1','j2'])
        >>> v = gp.Variable(m, "v", domain=[i, j])
        >>> v_t = v.t() # v_t is an ImplicitVariable
        >>> v_t_t = v_t.t() # you can get transpose of ImplicitVariable as well
        >>> v_t_t.domain
        [Set(name='i', domain=['*']), Set(name='j', domain=['*'])]

        """
        from gamspy.math.matrix import permute

        dims = [x for x in range(len(self.domain))]
        if len(dims) < 2:
            raise ValidationError(
                "Variable must contain at least 2 dimensions to transpose"
            )

        x = dims[-1]
        dims[-1] = dims[-2]
        dims[-2] = x
        return permute(self, dims)  # type: ignore

    @property
    def records(self) -> pd.DataFrame | float | None:
        if self.parent.records is None:
            return None

        recs = self.parent.records
        for idx, literal in self._scalar_domains:
            column_name = recs.columns[idx]
            recs = recs[recs[column_name] == literal]

        return recs

    def __neg__(self):
        return expression.Expression(None, "-", self)

    def __eq__(self, other):
        return expression.Expression(self, "=e=", other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def gamsRepr(self) -> str:
        representation = self.name
        domain = list(self.domain)
        if domain and self.permutation is not None:
            domain = utils._permute_domain(domain, self.permutation)

        for i, d in self._scalar_domains:
            domain.insert(i, d)

        if domain:
            representation += utils._get_domain_str(domain)

        return representation



from gamspy._symbols.implicits.implicit_equation import ImplicitEquation
from gamspy._symbols.implicits.implicit_parameter import ImplicitParameter
from gamspy._symbols.implicits.implicit_set import ImplicitSet
from gamspy._symbols.implicits.implicit_variable import ImplicitVariable

__all__ = [
    "ImplicitEquation",
    "ImplicitParameter",
    "ImplicitSet",
    "ImplicitVariable",
]



from __future__ import annotations

import gamspy as gp
import gamspy.formulations.nn.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.math import dim


class AvgPool2d:
    """
    Formulation generator for 2D Avg Pooling in GAMS.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    kernel_size : int | tuple[int, int]
        Filter size
    stride : int | tuple[int, int] | None
        Stride in the avg pooling, it is equal to kernel_size if not provided
    padding : int | tuple[int, int]
        Amount of padding to be added to input, by default 0
    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> from gamspy.math import dim
    >>> m = gp.Container()
    >>> # 2x2 avg pooling
    >>> ap1 = gp.formulations.AvgPool2d(m, (2, 2))
    >>> inp = gp.Variable(m, domain=dim((10, 1, 24, 24)))
    >>> out, eqs = ap1(inp)
    >>> type(out)
    <class 'gamspy._symbols.variable.Variable'>
    >>> [len(x) for x in out.domain]
    [10, 1, 12, 12]

    """

    def __init__(
        self,
        container: gp.Container,
        kernel_size: int | tuple[int, int],
        stride: int | tuple[int, int] | None = None,
        padding: int = 0,
        name_prefix: str | None = None,
    ):
        _kernel_size = utils._check_tuple_int(kernel_size, "kernel_size")
        if stride is None:
            stride = kernel_size

        _stride = utils._check_tuple_int(stride, "stride")
        _padding = utils._check_tuple_int(padding, "padding", allow_zero=True)

        self.container = container
        self.kernel_size = _kernel_size
        self.stride = _stride
        self.padding = _padding

        if name_prefix is None:
            name_prefix = gp.utils._get_unique_name()

        self._name_prefix = name_prefix

    def _set_bounds(
        self,
        input: gp.Parameter | gp.Variable,
        subset: gp.Set,
    ) -> tuple[gp.Parameter, gp.Parameter]:
        # Extract batch and channel dimensions from input domain
        N, C = input.domain[:2]
        H_out, W_out, Hf, Wf, H_in, W_in = subset.domain

        # Create subset2 mapping output positions to input positions
        subset2 = gp.Set(
            self.container,
            domain=[H_out, W_out, H_in, W_in],
            name=utils._generate_name(
                "s", self._name_prefix, "in_out_matching_2"
            ),
        )

        subset2[H_out, W_out, H_in, W_in] = gp.Sum(
            [Hf, Wf], subset[H_out, W_out, Hf, Wf, H_in, W_in]
        )

        # Initialize parameters for bounds
        lb = gp.Parameter(
            self.container,
            domain=[N, C, H_out, W_out],
            name=utils._generate_name("p", self._name_prefix, "output_lb"),
        )
        ub = gp.Parameter(
            self.container,
            domain=[N, C, H_out, W_out],
            name=utils._generate_name("p", self._name_prefix, "output_ub"),
        )

        # Calculate upper/lower bounds based on input type
        if isinstance(input, gp.Variable):
            # Use variable bounds if input is a variable
            ub[...] = gp.Smax(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input.up[N, C, H_in, W_in],
            )
            lb[...] = gp.Smin(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input.lo[N, C, H_in, W_in],
            )
        else:
            # Use parameter values directly if input is a parameter
            ub[...] = gp.Smax(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input[N, C, H_in, W_in],
            )
            lb[...] = gp.Smin(
                gp.Domain(H_in, W_in).where[subset2[H_out, W_out, H_in, W_in]],
                input[N, C, H_in, W_in],
            )

        # 0 padding on the edges can shift the minimum and max value
        kh, kw = self.kernel_size
        ph, pw = self.padding
        scale = ((kh - ph) * (kw - pw)) / (kh * kw)

        lb.where[lb > 0] = lb * scale
        ub.where[ub < 0] = ub * scale

        return lb, ub

    def __call__(
        self,
        input: gp.Parameter | gp.Variable,
        propagate_bounds: bool = True,
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Forward pass your input, generate output and equations required for
        calculating the average pooling. Unlike the min or max pooling avg
        pooling does not require binary variables or the big-M formulation.
        if propagate_bounds is True, it will also set the bounds for the
        output variable based on the input.
        Returns the output variable and the list of equations required for
        the avg pooling formulation.

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input to the max pooling 2d layer, must be in shape
                (batch x in_channels x height x width)
        propagate_bounds: bool
                If True, it will set the bounds for the output variable
                based on the input.
                Default value: True

        Returns
        -------
        tuple[gp.Variable, list[gp.Equation]]

        """
        if not isinstance(input, (gp.Parameter, gp.Variable)):
            raise ValidationError("Expected a parameter or a variable input")

        if not isinstance(propagate_bounds, bool):
            raise ValidationError("Expected a boolean for propagate_bounds")

        if len(input.domain) != 4:
            raise ValidationError(
                f"expected 4D input (got {len(input.domain)}D input)"
            )

        N, C_in, H_in, W_in = input.domain

        h_in = len(H_in)
        w_in = len(W_in)

        h_out, w_out = utils._calc_hw(
            self.padding, self.kernel_size, self.stride, h_in, w_in
        )

        out_var = gp.Variable(
            self.container,
            domain=dim([len(N), len(C_in), h_out, w_out]),
            name=utils._generate_name("v", self._name_prefix, "output"),
        )

        N, C, H_out, W_out = out_var.domain

        set_out = gp.Equation(
            self.container,
            domain=out_var.domain,
            name=utils._generate_name("e", self._name_prefix, "set_output"),
        )

        # expr must have domain N, C, H_out, W_out
        top_index = (
            (self.stride[0] * (gp.Ord(H_out) - 1)) - self.padding[0] + 1
        )
        left_index = (
            (self.stride[1] * (gp.Ord(W_out) - 1)) - self.padding[1] + 1
        )
        coeff = 1 / (self.kernel_size[0] * self.kernel_size[1])

        Hf, Wf = gp.math._generate_dims(self.container, self.kernel_size)
        Hf, Wf, H_in, W_in = utils._next_domains(
            [Hf, Wf, H_in, W_in], out_var.domain
        )

        subset = gp.Set(
            self.container,
            name=utils._generate_name(
                "s", self._name_prefix, "in_out_matching_1"
            ),
            domain=[H_out, W_out, Hf, Wf, H_in, W_in],
        )
        subset[
            H_out,
            W_out,
            Hf,
            Wf,
            H_in,
            W_in,
        ].where[
            (gp.Ord(H_in) == (top_index + gp.Ord(Hf) - 1))
            & (gp.Ord(W_in) == (left_index + gp.Ord(Wf) - 1))
        ] = True

        expr = gp.Sum(
            subset[H_out, W_out, Hf, Wf, H_in, W_in],
            input[N, C, H_in, W_in] * coeff,
        )

        set_out[...] = out_var[N, C, H_out, W_out] == expr

        # Set variable bounds if propagate_bounds is True
        if propagate_bounds:
            lb, ub = self._set_bounds(input, subset)
            out_var.lo[...] = lb
            out_var.up[...] = ub

        return out_var, [set_out]



from __future__ import annotations

import doctest
import glob
import json
import os
import shutil
import time
from os.path import join

import gamspy_base
import pytest

import gamspy as gp
from gamspy.exceptions import GamspyException, ValidationError


@pytest.mark.unit
def test_version():
    assert gp.__version__ == "1.10.0"


@pytest.mark.unit
def test_config():
    m = gp.Container()
    assert m.system_directory == gamspy_base.directory

    i = gp.Set(m, records=["i1", "i2"])
    a = gp.Parameter(m, domain=i, records=[("i1", 1), ("i2", 2)])
    with pytest.raises(ValidationError):  # GAMSPy catches the domain violation
        a["i3"] = 5

    gp.set_options({"DOMAIN_VALIDATION": 0})
    with pytest.raises(GamspyException):  # GAMS returns domain violation error
        a["i3"] = 5

    gp.set_options({"DOMAIN_VALIDATION": 1})

    e = gp.Equation(m, "e")
    model = gp.Model(m, "my_model", equations=[e])

    # no equation definition was found. ValidationError by default.
    with pytest.raises(ValidationError):
        model.solve()

    # no equation definition was found. Raises GamspyException because the validation is disabled.
    gp.set_options({"VALIDATION": 0})
    with pytest.raises(GamspyException):
        model.solve()

    gp.set_options({"VALIDATION": 1})


@pytest.mark.unit
def test_domain_checking_config_performance():
    gp.set_options({"DOMAIN_VALIDATION": 1})
    m = gp.Container()
    i = gp.Set(m, records=range(999))
    a = gp.Parameter(m, domain=i)
    start = time.time()
    for idx in range(999):
        _ = (
            a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
        )

    timing_with_validation = time.time() - start

    gp.set_options({"DOMAIN_VALIDATION": 0})
    m = gp.Container()
    i = gp.Set(m, records=range(999))
    a = gp.Parameter(m, domain=i)
    start = time.time()
    for idx in range(999):
        _ = (
            a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
            + a[idx]
        )
    timing_without_validation = time.time() - start
    gp.set_options({"DOMAIN_VALIDATION": 1})

    print(f"{timing_with_validation=}, {timing_without_validation=}")
    assert timing_without_validation < timing_with_validation, (
        f"{timing_with_validation=}, {timing_without_validation=}"
    )


def test_map_special_values():
    m = gp.Container()
    a = gp.Parameter(m, "a")
    a[...] = gp.SpecialValues.EPS
    assert a.getAssignment() == "a = EPS;"

    gp.set_options({"MAP_SPECIAL_VALUES": 0})
    m = gp.Container()
    a = gp.Parameter(m, "a")
    a[...] = gp.SpecialValues.EPS
    assert a.getAssignment() == "a = -0.0;"
    gp.set_options({"MAP_SPECIAL_VALUES": 1})


@pytest.mark.doc
def test_switcher():
    this = os.path.dirname(os.path.abspath(__file__))
    root = this.rsplit(os.sep, maxsplit=1)[0]
    with open(join(root, "docs", "_static", "switcher.json")) as file:
        switcher = json.loads(file.read())
        versions = [elem["version"] for elem in switcher]
        assert f"v{gp.__version__}" in versions


@pytest.fixture
def teardown():
    # Act and assert
    yield

    # Cleanup
    files = glob.glob("_*")
    for file in files:
        if os.path.isfile(file):
            os.remove(file)

    if os.path.exists("test"):
        shutil.rmtree("test")


@pytest.mark.doc
def test_docs():
    api_files = [
        file
        for file in glob.glob("src/**", recursive=True)
        if file.endswith(".py")
    ]

    for file in api_files:
        results = doctest.testfile(
            file,
            verbose=True,
            module_relative=False,
        )

        assert results.failed == 0



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._algebra.expression as expression
from gamspy.math.misc import MathOp

if TYPE_CHECKING:
    from gamspy._algebra.expression import Expression
    from gamspy._types import OperableType


def cos(x: OperableType) -> Expression:
    """
    Cosine of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import cos
    >>> import numpy as np
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = cos(np.pi)
    >>> r.toValue()
    np.float64(-1.0)

    """
    return expression.Expression(None, MathOp("cos", (x,)), None)


def cosh(x: OperableType) -> Expression:
    """
    Hyperbolic cosine of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import cosh
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = cosh(0)
    >>> r.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("cosh", (x,)), None)


def sin(x: OperableType) -> Expression:
    """
    Sine of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import sin
    >>> import numpy as np
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = sin(np.pi/2)
    >>> r.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("sin", (x,)), None)


def sinh(x: OperableType) -> Expression:
    """
    Hyperbolic sine of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import sinh
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = sinh(0)
    >>> r.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("sinh", (x,)), None)


def tan(x: OperableType) -> Expression:
    """
    Tangent of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import tan
    >>> import numpy as np
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = tan(np.pi/4)
    >>> round(r.toValue(), 2)
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("tan", (x,)), None)


def tanh(x: OperableType) -> Expression:
    """
    Hyperbolic tangent of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import tanh
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = tanh(0)
    >>> r.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("tanh", (x,)), None)


def acos(x: OperableType) -> Expression:
    """
    Inverse cosine of ``x``.

    Returns
    -------
    Expresion | float

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import acos
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = acos(1)
    >>> r.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("arccos", (x,)), None)


def asin(x: OperableType) -> Expression:
    """
    Inver sinus of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import asin
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = asin(0)
    >>> r.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("arcsin", (x,)), None)


def atan(x: OperableType) -> Expression:
    """
    Inverse tangent of ``x``.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import atan
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = atan(0)
    >>> r.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("arctan", (x,)), None)


def atan2(y: OperableType, x: OperableType) -> Expression:
    """
    Four-quadrant arctan function yielding ``arctan(y/x)``, which is the angle the vector ``(x,y)`` makes with ``(1,0)`` in radians.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import atan2
    >>> m = Container()
    >>> r = Parameter(m, "r")
    >>> r[...] = atan2(1,1)
    >>> r.toValue()
    np.float64(0.7853981633974483)

    """
    return expression.Expression(None, MathOp("arctan2", (y, x)), None)



import gamspy.formulations.nn as nn
import gamspy.formulations.piecewise as piecewise
from gamspy.formulations.nn import (
    AvgPool2d,
    Conv2d,
    Linear,
    MaxPool2d,
    MinPool2d,
)
from gamspy.formulations.piecewise import (
    pwl_convexity_formulation,
    pwl_interval_formulation,
)
from gamspy.formulations.shape import flatten_dims

__all__ = [
    "nn",
    "piecewise",
    "Conv2d",
    "MaxPool2d",
    "MinPool2d",
    "AvgPool2d",
    "Linear",
    "flatten_dims",
    "pwl_convexity_formulation",
    "pwl_interval_formulation",
]



from gamspy.formulations.nn.avgpool2d import AvgPool2d
from gamspy.formulations.nn.conv2d import Conv2d
from gamspy.formulations.nn.linear import Linear
from gamspy.formulations.nn.maxpool2d import MaxPool2d
from gamspy.formulations.nn.minpool2d import MinPool2d
from gamspy.formulations.nn.mpool2d import _MPool2d

__all__ = [
    "Conv2d",
    "_MPool2d",
    "MaxPool2d",
    "MinPool2d",
    "AvgPool2d",
    "Linear",
]



from __future__ import annotations

import math
import typing

import numpy as np

import gamspy as gp
from gamspy._symbols.implicits import (
    ImplicitVariable,
)
from gamspy.exceptions import ValidationError


def _generate_gray_code(n: int, n_bits: int) -> np.ndarray:
    """
    Returns an n x n_bits NumPy array containing gray codes.
    The bit difference between two consecutive rows is exactly
    1 bits. Required for the log piecewise linear formulation.
    """
    a = np.arange(n)
    b = a >> 1
    numbers = a ^ b
    numbers_in_bit_array = (
        (numbers[:, None] & (1 << np.arange(n_bits))) > 0
    ).astype(int)
    return numbers_in_bit_array


def _enforce_sos2_with_binary(lambda_var: gp.Variable) -> list[gp.Equation]:
    """
    Enforces SOS2 constraints using binary variables. This function is not suitable
    for generic SOS2 implementation since it restricts the lambda_var values to be
    between 0 and 1. However, it is usually faster than using SOS2 variables.

    Based on paper:
    `Modeling disjunctive constraints with a logarithmic number of binary variables and constraints
    <https://link.springer.com/article/10.1007/s10107-009-0295-4>`_
    """
    equations: list[gp.Equation] = []
    m = lambda_var.container
    count_x = len(lambda_var.domain[-1])
    # edge case
    lambda_var.lo[...] = 0
    lambda_var.up[...] = 1
    if count_x == 2:
        # if there are only 2 elements, it is already sos2
        return equations

    J = lambda_var.domain[-1]
    previous_domains = lambda_var.domain[:-1]

    l_len = math.ceil(math.log2(count_x - 1))
    I, L = gp.math._generate_dims(
        m,
        [
            count_x - 1,
            l_len,
        ],
    )

    J, I, L = gp.formulations.nn.utils._next_domains(
        [J, I, L], previous_domains
    )
    bin_var = m.addVariable(domain=[*previous_domains, L], type="binary")
    gray_code = _generate_gray_code(count_x - 1, l_len)

    B = m.addParameter(domain=[I, L], records=gray_code)

    JI = m.addSet(domain=[J, I])
    JI[J, I].where[(gp.Ord(J) == gp.Ord(I)) | (gp.Ord(J) - 1 == gp.Ord(I))] = 1

    use_set_1 = m.addSet(domain=[L, J])
    use_set_1[L, J].where[gp.Smin(JI[J, I], B[I, L]) == 1] = 1

    use_set_2 = m.addSet(domain=[L, J])
    use_set_2[L, J].where[gp.Smax(JI[J, I], B[I, L]) == 0] = 1

    sos2_eq_1 = m.addEquation(domain=[*previous_domains, L])
    sos2_eq_1[[*previous_domains, L]] = (
        gp.Sum(use_set_1[L, J], lambda_var[[*previous_domains, J]])
        <= bin_var[[*previous_domains, L]]
    )
    equations.append(sos2_eq_1)

    sos2_eq_2 = m.addEquation(domain=[*previous_domains, L])
    sos2_eq_2[[*previous_domains, L]] = (
        gp.Sum(use_set_2[L, J], lambda_var[[*previous_domains, J]])
        <= 1 - bin_var[[*previous_domains, L]]
    )
    equations.append(sos2_eq_2)

    return equations


def _enforce_discontinuity(
    lambda_var: gp.Variable,
    combined_indices: typing.Sequence[int],
) -> list[gp.Equation]:
    equations: list[gp.Equation] = []

    len_x_points = len(lambda_var.domain[-1])
    previous_domains = lambda_var.domain[:-1]

    m = lambda_var.container
    J, J2, SB = gp.math._generate_dims(
        m, [len_x_points, len_x_points, len(combined_indices)]
    )

    J, J2, SB = gp.formulations.nn.utils._next_domains(
        [J, J2, SB], previous_domains
    )

    di_param = [
        (str(i), str(j), str(j + 1)) for i, j in enumerate(combined_indices)
    ]

    select_set = m.addSet(domain=[SB, J, J2], records=di_param)
    select_var = m.addVariable(domain=[*previous_domains, SB], type="binary")

    select_equation = m.addEquation(domain=[*previous_domains, SB, J, J2])
    select_equation[[*previous_domains, select_set[SB, J, J2]]] = (
        lambda_var[[*previous_domains, J]]
        <= select_var[[*previous_domains, SB]]
    )
    equations.append(select_equation)

    select_equation_2 = m.addEquation(domain=[*previous_domains, SB, J, J2])
    select_equation_2[[*previous_domains, select_set[SB, J, J2]]] = (
        lambda_var[[*previous_domains, J2]]
        <= 1 - select_var[[*previous_domains, SB]]
    )
    equations.append(select_equation_2)

    return equations


def _check_points(
    x_points: typing.Sequence[int | float],
    y_points: typing.Sequence[int | float],
) -> tuple[list[int | float], list[int | float], list[int], list[int]]:
    return_x = []
    return_y = []
    discontinuous_indices = []
    none_indices = []

    if not isinstance(x_points, typing.Sequence):
        raise ValidationError("x_points are expected to be a sequence")

    if not isinstance(y_points, typing.Sequence):
        raise ValidationError("y_points are expected to be a sequence")

    if len(x_points) < 2:
        raise ValidationError(
            "piecewise linear functions require at least 2 points"
        )

    if len(y_points) != len(x_points):
        raise ValidationError("x_points and y_points have different lenghts")

    if x_points[0] is None or x_points[-1] is None:
        raise ValidationError("x_points cannot start or end with a None value")

    for x_p, y_p in zip(x_points, y_points):
        if (x_p is None and y_p is not None) or (
            x_p is not None and y_p is None
        ):
            raise ValidationError(
                "Both x and y must either be None or neither of them should be None"
            )

        if not isinstance(x_p, (float, int)) and x_p is not None:
            raise ValidationError("x_points contains non-numerical items")

        if not isinstance(y_p, (float, int)) and y_p is not None:
            raise ValidationError("y_points contains non-numerical items")

    for i in range(len(x_points) - 1):
        if x_points[i] is None and x_points[i + 1] is None:
            raise ValidationError(
                "x_points cannot contain two consecutive None values"
            )

        if x_points[i] is None and x_points[i - 1] >= x_points[i + 1]:
            raise ValidationError(
                "A value following a None must be strictly greater than the value preceding the None"
            )

        if (
            (x_points[i] is not None)
            and (x_points[i + 1] is not None)
            and (x_points[i + 1] < x_points[i])
        ):
            raise ValidationError(
                "x_points should be in an non-decreasing order"
            )

        if x_points[i] is not None:
            return_x.append(x_points[i])
            return_y.append(y_points[i])

        if x_points[i] == x_points[i + 1]:
            discontinuous_indices.append(len(return_x) - 1)
        elif x_points[i] is None:
            none_indices.append(len(return_x) - 1)

    return_x.append(x_points[-1])
    return_y.append(y_points[-1])
    return return_x, return_y, discontinuous_indices, none_indices


def _indicator(
    indicator_var: gp.Variable,
    indicator_val: typing.Literal[0, 1],
    expr: gp.Expression,
) -> list[gp.Equation]:
    # We will make this generic and public
    if not isinstance(indicator_var, (gp.Variable, ImplicitVariable)):
        raise ValidationError("indicator_var needs to be a variable")

    if indicator_var.type != "binary":
        raise ValidationError("indicator_var needs to be a binary variable")

    if indicator_val != 0 and indicator_val != 1:
        raise ValidationError("indicator_val needs to be 1 or 0")

    if not isinstance(expr, gp.Expression):
        raise ValidationError("expr needs to be an expression")

    if expr.data not in {"=l=", "=e=", "=g="}:
        raise ValidationError("expr needs to be inequality or equality")

    if len(expr.domain) != len(indicator_var.domain):
        raise ValidationError(
            "indicator_var and expr must have the same domain"
        )

    for i in range(len(expr.domain)):
        if expr.domain[i].name != indicator_var.domain[i].name:
            raise ValidationError(
                "indicator_var and expr must have the same domain"
            )

    if expr.data == "=e=":
        # sos1(bin_var, lhs - rhs) might be better
        eqs1 = _indicator(
            indicator_var,
            indicator_val,
            expr.left <= expr.right,  # type: ignore
        )
        eqs2 = _indicator(
            indicator_var,
            indicator_val,
            -expr.left <= -expr.right,  # type: ignore
        )
        return [*eqs1, *eqs2]

    if expr.data == "=g=":
        return _indicator(
            indicator_var,
            indicator_val,
            -expr.left <= -expr.right,  # type: ignore
        )

    equations = []
    m = indicator_var.container

    slack_var = m.addVariable(domain=expr.domain, type="positive")
    slack_eq = m.addEquation(
        domain=expr.domain, definition=(expr.left - slack_var <= expr.right)
    )
    equations.append(slack_eq)

    expr_domain = ... if len(expr.domain) == 0 else [*expr.domain]

    sos_dim = gp.math._generate_dims(m, [2])[0]
    sos1_var = m.addVariable(domain=[*expr.domain, sos_dim], type="sos1")
    sos1_eq_1 = m.addEquation(domain=expr.domain)
    if indicator_val == 1:
        sos1_eq_1[...] = (
            sos1_var[[*expr.domain, "0"]] == indicator_var[expr_domain]
        )
    else:
        sos1_eq_1[...] = (
            sos1_var[[*expr.domain, "0"]] == 1 - indicator_var[expr_domain]
        )
    equations.append(sos1_eq_1)

    sos1_eq_2 = m.addEquation(domain=expr.domain)
    sos1_eq_2[...] = sos1_var[[*expr.domain, "1"]] == slack_var[expr_domain]
    equations.append(sos1_eq_2)

    return equations


def _generate_ray(
    container: gp.Container, domain: list[gp.Set]
) -> tuple[gp.Variable, gp.Variable, list[gp.Equation]]:
    # if b_var == 0 => x_var = 0 o.w x_var >= 0
    # effectively x_var <= bigM * b_var without bigM
    x_var = container.addVariable(domain=domain, type="positive")
    b_var = container.addVariable(domain=domain, type="binary")
    eqs = _indicator(b_var, 0, x_var <= 0)
    return x_var, b_var, eqs


def points_to_intervals(
    x_points: typing.Sequence[int | float],
    y_points: typing.Sequence[int | float],
    discontinuous_points: typing.Sequence[int],
) -> list[tuple[int | float, int | float, int | float, int | float]]:
    result: list[
        tuple[int | float, int | float, int | float, int | float]
    ] = []
    finished_at_disc = True
    for i in range(len(x_points) - 1):
        x1 = x_points[i]
        x2 = x_points[i + 1]
        y1 = y_points[i]
        y2 = y_points[i + 1]

        if i in discontinuous_points:
            if finished_at_disc:
                result.append((x1, x1, 0, y1))

            finished_at_disc = True
        else:
            finished_at_disc = False
            slope = (y2 - y1) / (x2 - x1)
            offset = y1 - (slope * x1)
            result.append((x1, x2, slope, offset))

    if finished_at_disc:
        result.append((x2, x2, 0, y2))

    return result


def _get_end_slopes(
    x_points: typing.Sequence[int | float],
    y_points: typing.Sequence[int | float],
) -> tuple[float, float]:
    if x_points[-1] != x_points[-2]:
        m_pos = (y_points[-1] - y_points[-2]) / (x_points[-1] - x_points[-2])
    else:
        m_pos = 0

    if x_points[0] != x_points[1]:
        m_neg = (y_points[0] - y_points[1]) / (x_points[0] - x_points[1])
    else:
        m_neg = 0

    return m_neg, m_pos


def pwl_interval_formulation(
    input_x: gp.Variable,
    x_points: typing.Sequence[int | float],
    y_points: typing.Sequence[int | float],
    bound_left: bool = True,
    bound_right: bool = True,
) -> tuple[gp.Variable, list[gp.Equation]]:
    """
    This function implements a piecewise linear function using the intervals formulation.
    Given an input (independent) variable `input_x`, along with the defining `x_points`
    and corresponding `y_points` of the piecewise function, it constructs the dependent
    variable `y` and formulates the equations necessary to define the function.

    Here is the interval formulation:

    .. math::
        \\lambda_i \\geq b_i * LB_i \\quad \\forall{i}

        \\lambda_i \\leq b_i * UB_i \\quad \\forall{i}

        \\sum_{i}{b_i} = 1

        x = \\sum_{i}{\\lambda_i}

        y = \\sum_{i}{(\\lambda_i * slope_i) + (b_i * offset_i) }

        b_i \\in \\{0, 1\\} \\quad \\forall{i}

    The implementation handles discontinuities in the function. To represent a
    discontinuity at a specific point `x_i`, include `x_i` twice in the `x_points`
    array with corresponding values in `y_points`. For example, if `x_points` =
    [1, 3, 3, 5] and `y_points` = [10, 30, 50, 70], the function allows y to take
    either 30 or 50 when x = 3. Note that discontinuities introduce additional
    binary variables.

    It is possible to disallow a specific range by including `None` in both
    `x_points` and the corresponding `y_points`. For example, with
    `x_points` = `[1, 3, None, 5, 7]` and `y_points` = `[10, 35, None, -20, 40]`,
    the range between 3 and 5 is disallowed for `input_x`.

    However, `x_points` cannot start or end with a `None` value, and a `None`
    value cannot be followed by another `None`. Additionally, if `x_i` is `None`,
    then `y_i` must also be `None`. Similar to the discontinuities, disallowed
    ranges always introduce additional binary variables.

    The input variable `input_x` is restricted to the range defined by
    `x_points` unless `bound_left` or `bound_right` is set to False. Setting
    either to True, creates SOS1 type of variables. When `input_x` is not bound,
    you can assume as if the first and/or the last line segments are extended.

    Returns the dependent variable `y` and the equations required to model the
    piecewise linear relationship.

    Parameters
    ----------
    x : gp.Variable
        Independent variable of the piecewise linear function
    x_points: typing.Sequence[int | float]
        Break points of the piecewise linear function in the x-axis
    y_points: typing.Sequence[int | float]
        Break points of the piecewise linear function in the y-axis
    bound_left: bool = True
        If input_x should be limited to start from x_points[0]
    bound_right: bool = True
        If input_x should be limited to end at x_points[-1]

    Returns
    -------
    tuple[gp.Variable, list[Equation]]

    Examples
    --------
    >>> from gamspy import Container, Variable, Set
    >>> from gamspy.formulations import pwl_interval_formulation
    >>> m = Container()
    >>> x = Variable(m, "x")
    >>> y, eqs = pwl_interval_formulation(x, [-1, 4, 10, 10, 20], [-2, 8, 15, 17, 37])

    """

    if not isinstance(input_x, gp.Variable):
        raise ValidationError("input_x is expected to be a Variable")

    if not isinstance(bound_left, bool):
        raise ValidationError("bound_left is expected to be a boolean")

    if not isinstance(bound_right, bool):
        raise ValidationError("bound_right is expected to be a boolean")

    x_points, y_points, discontinuous_indices, none_indices = _check_points(
        x_points, y_points
    )
    combined_indices = list({*discontinuous_indices, *none_indices})
    equations = []

    intervals = points_to_intervals(x_points, y_points, combined_indices)
    lowerbounds_input = []
    upperbounds_input = []
    slopes_input = []
    offsets_input = []
    for i, (lb, ub, slope, offset) in enumerate(intervals):
        lowerbounds_input.append((str(i), lb))
        upperbounds_input.append((str(i), ub))
        slopes_input.append((str(i), slope))
        offsets_input.append((str(i), offset))

    input_domain = input_x.domain
    m = input_x.container

    J = gp.math._generate_dims(m, [len(intervals)])[0]
    J = gp.formulations.nn.utils._next_domains([J], input_domain)[0]

    lowerbounds = m.addParameter(domain=[J], records=lowerbounds_input)
    upperbounds = m.addParameter(domain=[J], records=upperbounds_input)
    slopes = m.addParameter(domain=[J], records=slopes_input)
    offsets = m.addParameter(domain=[J], records=offsets_input)
    bin_var = m.addVariable(domain=[*input_domain, J], type="binary")

    lambda_var = m.addVariable(domain=[*input_domain, J])

    set_lambda_lowerbound = m.addEquation(domain=lambda_var.domain)
    set_lambda_lowerbound[...] = lowerbounds * bin_var <= lambda_var
    equations.append(set_lambda_lowerbound)

    set_lambda_upperbound = m.addEquation(domain=lambda_var.domain)
    set_lambda_upperbound[...] = upperbounds * bin_var >= lambda_var
    equations.append(set_lambda_upperbound)

    out_y = m.addVariable(domain=input_domain)

    x_term = 0
    y_term = 0
    pick_one_term = 0

    if bound_left is False or bound_right is False:
        m_neg, m_pos = _get_end_slopes(x_points, y_points)

    if bound_left:
        out_y.lo[...] = min(y_points)
    else:
        x_neg_inf, b_neg_inf, eqs_neg_inf = _generate_ray(m, input_domain)
        equations.extend(eqs_neg_inf)

        x_term += -x_neg_inf + (b_neg_inf * x_points[0])
        y_term += -(m_neg * x_neg_inf) + (b_neg_inf * y_points[0])
        pick_one_term += b_neg_inf

    if bound_right:
        out_y.up[...] = max(y_points)
    else:
        x_pos_inf, b_pos_inf, eqs_pos_inf = _generate_ray(m, input_domain)
        equations.extend(eqs_pos_inf)

        x_term += x_pos_inf + (b_pos_inf * x_points[-1])
        y_term += (m_pos * x_pos_inf) + (b_pos_inf * y_points[-1])
        pick_one_term += b_pos_inf

    pick_one = m.addEquation(domain=input_domain)
    pick_one[...] = gp.Sum(J, bin_var) + pick_one_term == 1
    equations.append(pick_one)

    set_x = m.addEquation(domain=input_domain)
    set_x[...] = input_x == gp.Sum(J, lambda_var) + x_term
    equations.append(set_x)

    set_y = m.addEquation(domain=input_domain)
    set_y[...] = (
        out_y
        == gp.Sum(J, lambda_var * slopes)
        + gp.Sum(J, bin_var * offsets)
        + y_term
    )
    equations.append(set_y)

    return out_y, equations


def pwl_convexity_formulation(
    input_x: gp.Variable,
    x_points: typing.Sequence[int | float],
    y_points: typing.Sequence[int | float],
    using: typing.Literal["binary", "sos2"] = "binary",
    bound_left: bool = True,
    bound_right: bool = True,
) -> tuple[gp.Variable, list[gp.Equation]]:
    """
    This function implements a piecewise linear function using the convexity formulation.
    Given an input (independent) variable `input_x`, along with the defining `x_points`
    and corresponding `y_points` of the piecewise function, it constructs the dependent
    variable `y` and formulates the equations necessary to define the function.

    Here is the convexity formulation:

    .. math::
        x = \\sum_{i}{x\\_points_i * \\lambda_i}

        y = \\sum_{i}{y\\_points_i * \\lambda_i}

        \\sum_{i}{\\lambda_i} = 1

        \\lambda_i \\in SOS2


    By default, SOS2 variables are implemented using binary variables.
    See
    `Modeling disjunctive constraints with a logarithmic number of binary variables and constraints
    <https://link.springer.com/article/10.1007/s10107-009-0295-4>`_
    . However, you can switch to SOS2 (Special Ordered Set Type 2) by setting the
    `using` parameter to `"sos2"`.

    The implementation handles discontinuities in the function. To represent a
    discontinuity at a specific point `x_i`, include `x_i` twice in the `x_points`
    array with corresponding values in `y_points`. For example, if `x_points` =
    [1, 3, 3, 5] and `y_points` = [10, 30, 50, 70], the function allows y to take
    either 30 or 50 when x = 3. Note that discontinuities always introduce
    additional binary variables, regardless of the value of the using argument.

    It is possible to disallow a specific range by including `None` in both
    `x_points` and the corresponding `y_points`. For example, with
    `x_points` = `[1, 3, None, 5, 7]` and `y_points` = `[10, 35, None, -20, 40]`,
    the range between 3 and 5 is disallowed for `input_x`.

    However, `x_points` cannot start or end with a `None` value, and a `None`
    value cannot be followed by another `None`. Additionally, if `x_i` is `None`,
    then `y_i` must also be `None`. Similar to the discontinuities, disallowed
    ranges always introduce additional binary variables, regardless of the value
    of the using argument.

    The input variable `input_x` is restricted to the range defined by
    `x_points` unless `bound_left` or `bound_right` is set to False. Setting
    either to True, creates SOS1 type of variables. When `input_x` is not bound,
    you can assume as if the first and/or the last line segments are extended.

    Returns the dependent variable `y` and the equations required to model the
    piecewise linear relationship.

    Parameters
    ----------
    x : gp.Variable
        Independent variable of the piecewise linear function
    x_points: typing.Sequence[int | float]
        Break points of the piecewise linear function in the x-axis
    y_points: typing.Sequence[int| float]
        Break points of the piecewise linear function in the y-axis
    using: str = "binary"
        What type of variable is used during implementing piecewise function
    bound_left: bool = True
        If input_x should be limited to start from x_points[0]
    bound_right: bool = True
        If input_x should be limited to end at x_points[-1]

    Returns
    -------
    tuple[gp.Variable, list[Equation]]

    Examples
    --------
    >>> from gamspy import Container, Variable, Set
    >>> from gamspy.formulations import pwl_convexity_formulation
    >>> m = Container()
    >>> x = Variable(m, "x")
    >>> y, eqs = pwl_convexity_formulation(x, [-1, 4, 10, 10, 20], [-2, 8, 15, 17, 37])

    """
    if using not in {"binary", "sos2"}:
        raise ValidationError(
            "Invalid value for the using argument."
            "Possible values are 'binary' and 'sos2'"
        )

    if not isinstance(input_x, gp.Variable):
        raise ValidationError("input_x is expected to be a Variable")

    if not isinstance(bound_left, bool):
        raise ValidationError("bound_left is expected to be a boolean")

    if not isinstance(bound_right, bool):
        raise ValidationError("bound_right is expected to be a boolean")

    x_points, y_points, discontinuous_indices, none_indices = _check_points(
        x_points, y_points
    )
    combined_indices = list({*discontinuous_indices, *none_indices})

    m = input_x.container
    input_domain = input_x.domain
    out_y = m.addVariable(domain=input_domain)
    equations = []

    J = gp.math._generate_dims(m, [len(x_points)])[0]
    J = gp.formulations.nn.utils._next_domains([J], input_domain)[0]

    x_par = m.addParameter(domain=[J], records=np.array(x_points))
    y_par = m.addParameter(domain=[J], records=np.array(y_points))

    lambda_var = m.addVariable(
        domain=[*input_domain, J], type="free" if using == "binary" else "sos2"
    )

    lambda_var.lo[...] = 0
    lambda_var.up[...] = 1

    x_term = 0
    y_term = 0

    if bound_left is False or bound_right is False:
        m_neg, m_pos = _get_end_slopes(x_points, y_points)

    if bound_left:
        out_y.lo[...] = min(y_points)
    else:
        x_neg_inf, b_neg_inf, eqs_neg_inf = _generate_ray(m, input_domain)
        equations.extend(eqs_neg_inf)

        limit_b_neg_inf = m.addEquation(domain=b_neg_inf.domain)
        limit_b_neg_inf[...] = b_neg_inf <= lambda_var[[*input_domain, "0"]]
        equations.append(limit_b_neg_inf)

        x_term += -x_neg_inf
        y_term += -(m_neg * x_neg_inf)

    if bound_right:
        out_y.up[...] = max(y_points)
    else:
        x_pos_inf, b_pos_inf, eqs_pos_inf = _generate_ray(m, input_domain)
        equations.extend(eqs_pos_inf)

        limit_b_pos_inf = m.addEquation(domain=b_pos_inf.domain)
        last = str(len(J) - 1)
        limit_b_pos_inf[...] = b_pos_inf <= lambda_var[[*input_domain, last]]
        equations.append(limit_b_pos_inf)

        x_term += x_pos_inf
        y_term += m_pos * x_pos_inf

    lambda_sum = m.addEquation(domain=input_x.domain)
    lambda_sum[...] = gp.Sum(J, lambda_var) == 1
    equations.append(lambda_sum)

    set_x = m.addEquation(domain=input_x.domain)
    set_x[...] = input_x == gp.Sum(J, x_par * lambda_var) + x_term
    equations.append(set_x)

    set_y = m.addEquation(domain=input_x.domain)
    set_y[...] = out_y == gp.Sum(J, y_par * lambda_var) + y_term
    equations.append(set_y)

    if using == "binary":
        extra_eqs = _enforce_sos2_with_binary(lambda_var)
        equations.extend(extra_eqs)

    if len(combined_indices) > 0:
        extra_eqs = _enforce_discontinuity(lambda_var, combined_indices)
        equations.extend(extra_eqs)

    return out_y, equations
